import numpy as np
import os
from scipy.spatial.transform import Rotation as R

class BVHParser:
    def __init__(self, filepath):
        self.filepath = filepath
        self.header = []
        self.motion = []
        self.load()
    
    def load(self):
        if not os.path.exists(self.filepath):
            raise FileNotFoundError(f"Error: The file '{self.filepath}' was not found. Please check the path.")
        
        with open(self.filepath, 'r') as f:
            lines = f.readlines()
        
        try:
            motion_index = lines.index("MOTION\n") + 3  # Skip 'MOTION' and first two lines
            self.header = lines[:motion_index]
            self.motion = [list(map(float, line.strip().split())) for line in lines[motion_index:]]
        except ValueError:
            raise ValueError("Error: Invalid BVH file format. 'MOTION' section not found.")
    
    def save(self, output_path):
        with open(output_path, 'w') as f:
            f.writelines(self.header)
            for frame in self.motion:
                f.write(" ".join(map(str, frame)) + "\n")

def smooth_loop_bvh(input_bvh, output_bvh, blend_frames=30):
    if not os.path.exists(input_bvh):
        raise FileNotFoundError(f"Error: The input file '{input_bvh}' does not exist.")
    
    # Load BVH file
    bvh = BVHParser(input_bvh)
    motion = np.array(bvh.motion)  # Shape: (num_frames, num_channels)
    
    num_frames = motion.shape[0]
    if num_frames < blend_frames:
        raise ValueError("Not enough frames for blending.")
    
    # Compute cyclic offset (difference between last and first frame)
    offset = motion[-1] - motion[0]
    
    # Gradually adjust all frames to distribute the transition smoothly
    for i in range(num_frames):
        alpha = i / num_frames  # Smooth transition factor across the whole sequence
        motion[i] -= alpha * offset  # Apply gradual offset correction
    
    # Save the modified BVH file
    bvh.motion = motion.tolist()
    bvh.save(output_bvh)
    print(f"Looped BVH saved to: {output_bvh}")

input_path = "Bear.bvh"
output_path = "Bear_Smoothed.bvh"

if os.path.exists(input_path):
    smooth_loop_bvh(input_path, output_path)
else:
    print(f"Error: The file '{input_path}' was not found. Please check the path and try again.")
