#define GL_DO_NOT_WARN_IF_MULTI_GL_VERSION_HEADERS_INCLUDED

#import <Foundation/Foundation.h>
#import <MeshP2/MeshP2.h>
#import <vector>
#import "Star.h"

#define NEAR_CLIPPING_PLANE 1.0
#define FAR_CLIPPING_PLANE  10000.0

@interface Okudagram : NSObject {					                        // In honour of Michael & Denise Okuda - Technical Consultants on Star Trek: The Next Generation

    bool             P2_Logging_Enabled;

    NSTimeInterval   StartTime;                                             // Stores time when Okudagram was instantiated; needed for time-based shader effects

    NSMutableArray	*Star_Array;					                        // Array of Star objects traversing the surface of an ellipsoid i.e. 'warp bubble'

    glm::vec2        mouseCoords;                                           // Holds the current normalized mouse coordinates
    GLfloat          frustum[6][4];                                         // Holds the current view frustum plane equations
    GLfloat          warpSpeed;                                             // Current starfield warp speed setting
    GLboolean        wireframeMode;                                         // Wireframe polygons or solid fill?
    GLboolean        skymapVisible;                                         // Deep field Skymap visible or not?
    GLboolean        starfieldVisible;                                      // Animated Starfield visible?
    GLboolean        warpfieldVisible;                                      // Animated Warpfield visible?
    GLboolean        oceanVisible;                                          // Ocean wave effects visible?
    GLboolean        hyperspaceLocked;                                      // Hyperspace effect locked?

    GLuint	         Skymap_TextureID;                                      // Skymap texture, shader and program IDs...
    GLuint           Skymap_VertexShader;
    GLuint           Skymap_FragmentShader;
    GLuint           Skymap_ShaderProgram;
    
    GLuint           Starfield_TextureID;                                   // Starfield texture, shader and program IDs...
    GLuint           Starfield_VertexShader;
    GLuint           Starfield_FragmentShader;
    GLuint           Starfield_ShaderProgram;
    
    GLuint           Warpfield_VertexShader;                                // Warpfield shader and program IDs...
    GLuint           Warpfield_FragmentShader;
    GLuint           Warpfield_ShaderProgram;
    
    GLuint           Ocean_VertexShader;                                    // Ocean shader and program IDs...
    GLuint           Ocean_FragmentShader;
    GLuint           Ocean_ShaderProgram;
    
    GLuint           SkymapVAO;                                             // Vertex Array Objects
    GLuint           StarfieldVAO;
    GLuint           WarpfieldVAO;                                          // (VBO not needed as Warpfield.vert generates all vertex attributes)
    GLuint           OceanVAO;                                              // (VBO not needed as Ocean.vert     generates all vertex attributes)

    GLuint           SkymapVBO;                                             // Vertex Buffer Objects
    GLuint           StarfieldVBO;

    Matrix4f        *ModelViewMatrix;                                       // ModelView matrix, updated when the okudagram is rotated, scaled & zoomed; sent to vertex shader
    Matrix4f        *ProjectionMatrix;                                      // Projection matrix, updated when the okudagram is loaded, or window reshaped; sent to the shader
    
    GLfloat          globalLightConstantAttenuation;
    GLfloat          globalLightLinearAttenuation;
    GLfloat          globalLightQuadraticAttenuation;
    
    GLint            locModelViewMatrix;                                    // Vertex/Fragment Shader Uniform Locations...
    GLint            locProjectionMatrix;
    GLint            loc2ndModelViewMatrix;                                 // 2nd set of MVP matrix uniform locations are needed...
    GLint            loc2ndProjectionMatrix;                                // ... as uniforms aren't shared between shader programs
    GLint            locCubeMapFlag;
    GLint            locCubeMapUnit;
    GLint            locTransformMatrix;
    GLint            locAttenuatedColor;
    GLint            locDeltaTime;                                          // Time since Okudagram was instantiated; needed for time-based shader effects
    GLint            loc2ndDeltaTime;                                       // 2nd Okudagram instantiation time (uniforms aren't shared between shaders)
    GLint            locMouseCoords;                                        // Normalized mouse coordinates, updated on otherMouseDragged
}

-(id)		init  :(GLfloat)in_Warp  :(int)in_Density  :(Vector3f*)in_MeshAxisLimits  :(bool)in_Logging;

-(void)     SetModelViewMatrix  :(Matrix4f*)in_ModelView_Matrix;
-(void)     SetProjectionMatrix :(Matrix4f*)in_Projection_Matrix;
-(void)     SetMouseCoords      :(glm::vec2)in_MouseCoords;

-(GLfloat)  GetWarpSpeed;
-(void)     SetWarpSpeed        :(GLfloat)in_WarpSpeed;

-(void)     SetSkymap           :(int)in_SkymapIndex;

-(GLboolean)GetHyperspaceLocked;
-(void)     SetHyperspaceLocked :(GLboolean)in_HyperspaceLocked;

-(void)     ToggleWireframeMode;
-(void)     ToggleSkymap;
-(void)     ToggleStarfield;
-(void)     ToggleWarpfield;
-(void)     ToggleOcean;

-(void)     SetupVAOsVBOs;
-(void)     SetupShaders;
-(GLuint)   CompileShaderType   :(GLenum)type  :(NSString*)file;
-(void)     ShaderStatus        :(GLuint)Object  :(GLenum)Parameter;

-(void)     LoadSkymap;                                                             // Loads a deep field textured Cubemap
-(void) 	Display;                                                                // Display the Okudagram on the screen
-(void)		DrawSkymap;						                                        // Draws deep field skymap as textures
-(void) 	DrawStarfield;													        // Draws 3D animated starfield (warp drive effect)

-(void)     ExtractFrustum;                                                         // Extracts the viewing frustum (i.e. field of view)
-(bool)     PointInFrustum      :(float)x  :(float)y  :(float)z;                    // Determine if a point is within viewing frustum...
-(bool)     SphereInFrustum     :(float)x  :(float)y  :(float)z  :(float)radius;    // ... or a sphere...
-(bool)     CubeInFrustum       :(float)x  :(float)y  :(float)z  :(float)size;      // ... or a cube.

@end
