#import "Star.h"

@implementation Star

-(id) init  :(int)in_WarpSpeed  :(Vector3f*)in_AxisLimits {
    
    if    ((self = [super init])) {
        
        warpSpeed   = in_WarpSpeed;
        maxDisks    = fmin (in_AxisLimits.getZ * 1000.0, FAR_CLIPPING_PLANE);       // Set maximum ellipsoid length as a multiple of mesh's length
        maxEdges    = fmax (in_AxisLimits.getX * in_AxisLimits.getY, 360.0);        // Fudge edge number; the segment around the ellipsoid surface
        disk        = 0.0;
        
        [self reset  :true  :false];                                                // Reset all stars (with the Hyperspace drive effect disabled)
    }
    
    return self;
}

-(void) reset  :(bool)in_InitAll  :(bool)in_HyperspaceEffect {
    
/*
    Reset the star's position by building and saving a transform matrix from the randomized disk, edge, deltaZ and rollAngle values computed here.
    DrawStarfield will use both disk and edge to update the star's ellipsoid coordinates, saving them to the translation section of the transform.
    DrawStarfield then uploads the transform to the vertex shader (via a uniform location) to translate, rotate and scale the star's VBO vertices.
*/

    if (in_InitAll) {                                                               // Reset both disk (pole-to-pole Z axis) & edge (X and Y axes)
        
        disk    = (disk == 0.0) ? SSRandomFloatBetween (1.0, maxDisks) : maxDisks;  // On 1st frame display, randomize disk. Otherwise, just reset
        edge    = SSRandomFloatBetween (1.0, maxEdges);
    }

    if (in_HyperspaceEffect) {                                                      // Enable Hyperspace effect, but only on warp speed changes

        hyperspace      = in_HyperspaceEffect;                                      // Reset Hyperspace effect flag for Okudagram DrawStarfield
        hyperspaceIndex = 0;                                                        // Reset Hyperspace index, the count of accumulated effects
    }
    
    warpSpeed   = fmin (warpSpeed, MAX_SPEED);                                      // Set some limits to ensure warpSpeed maintains reasonable values
    warpSpeed   = fmax (warpSpeed, MIN_SPEED);                                      // (Helps avoid oddities, for example when '-' key hit repeatedly)

    float zAxisScaling = warpSpeed * SSRandomFloatBetween (warpSpeed, MAX_SPEED);   // Fudge an additional random element to enhance starfield visuals

    glm::vec3   scaleVector (1.0, 1.0, zAxisScaling);                               // Stretch out star along the Z-axis to give the illusion of speed

    deltaZ      = sqrt (scaleVector.z);                                             // Use the Z-axis stretch factor as the star's drawtime step rate

    rollAngle   = glm::radians (edge * 360.0 / maxEdges);                           // Convert modulo 360° edge number to an ellipsoid surface angle
    rollAngle   = glm::tan     (rollAngle);                                         // Get the angle which places the star flat on ellipsoid surface
    rollAngle   = glm::atan    (rollAngle);                                         // Cancel any quadrant effects to finalize the star's roll angle

    transform   = glm::mat4    (1.0f);                                                       // From Identity, build transform in OpenGL TRS order:
    transform   = glm::scale   (transform, scaleVector);                                     // Scale  vertex on Z axis as a function of warp speed
    transform   = glm::rotate  (transform, rollAngle, glm::vec3 (0.0, 0.0, 1.0));            // Rotate vertex on Z axis to flatten on ellipsoid surface
    transform   = glm::rotate  (transform, glm::radians (90.0f), glm::vec3 (0.0, 1.0, 0.0)); // Rotate vertex on Y axis to align with ellipsoid's poles
}

-(GLfloat)      getX                { return x; }
-(GLfloat)      getY                { return y; }
-(GLfloat)      getZ                { return z; }
-(GLfloat)	    getDisk			    { return disk; }
-(GLfloat)	    getEdge			    { return edge; }
-(GLfloat)	    getDeltaZ		    { return deltaZ; }
-(GLfloat)      getPhi              { return phi; }
-(GLfloat)      getTheta            { return theta; }
-(GLfloat)	    getRollAngle	    { return rollAngle; }
-(GLfloat)      getMaxDisks         { return maxDisks; }
-(GLfloat)      getMaxEdges         { return maxEdges; }
-(GLboolean)    getHyperspace       { return hyperspace; }
-(glm::mat4)    getTransform        { return transform; }
-(glm::vec4)    getColor            { return color; }

-(void)         setX                :(GLfloat)newx              { x = newx; }
-(void)         setY                :(GLfloat)newy              { y = newy; }
-(void)         setZ                :(GLfloat)newz              { z = newz; }
-(void)         setPhi              :(GLfloat)newphi            { phi = newphi; }
-(void)         setTheta            :(GLfloat)newtheta          { theta = newtheta; }
-(void)		    setDisk             :(GLfloat)newdisk           { disk = newdisk; }
-(void)		    setEdge             :(GLfloat)newedge	        { edge = newedge; }
-(void)         setColor            :(glm::vec4)newColor        { color = newColor; }
-(void)         setTransform        :(glm::mat4)newTransform    { transform = newTransform; }
-(void)         setWarpSpeed        :(GLfloat)newWarpSpeed      { warpSpeed = newWarpSpeed; }
-(void)         setHyperspace       :(GLboolean)newHyperspace   { hyperspace = newHyperspace; }

-(GLint)        incHyperspaceIndex  { return hyperspaceIndex++; }

@end
